package chocopy.pa3;

import java.nio.charset.Charset;
import java.nio.charset.StandardCharsets;
import java.util.*;

import chocopy.common.analysis.SymbolTable;
import chocopy.common.analysis.AbstractNodeAnalyzer;
import chocopy.common.analysis.types.SymbolType;
import chocopy.common.analysis.types.ValueType;
import chocopy.common.astnodes.*;
import chocopy.common.codegen.*;

import static chocopy.common.codegen.RiscVBackend.Register.*;

/**
 * This is where the main implementation of PA3 will live.
 *
 * A large part of the functionality has already been implemented
 * in the base class, CodeGenBase. Make sure to read through that
 * class, since you will want to use many of its fields
 * and utility methods in this class when emitting code.
 *
 * Also read the PDF spec for details on what the base class does and
 * what APIs it exposes for its sub-class (this one). Of particular
 * importance is knowing what all the SymbolInfo classes contain.
 */


public class CodeGenImpl extends CodeGenBase {
    private class BetterRsicVBackend {
        private final RiscVBackend backend;
        BetterRsicVBackend(RiscVBackend _backend) {
            backend = _backend;
        }
        public void emitNoop(String comment) {
            if(comment != null) {
                backend.emitMV(ZERO, ZERO, "Noop: " + comment);
            }
        }
        public void emitPush(RiscVBackend.Register reg, String comment) {
            backend.emitADDI(SP, SP, -1 * backend.getWordSize(), comment);
            backend.emitSW(reg, SP, 0, comment);
        }
        public void emitPop(RiscVBackend.Register reg, String comment) {
            if(reg != null)
                backend.emitLW(reg, SP, 0, comment);
            backend.emitADDI(SP, SP, backend.getWordSize(), comment);
        }
        public void emitCall(Label calledLabel, String comment) {
            // Arguments should be already pushed to stack.
            emitNoop(comment);
            emitPush(RA, "backup return address of parent.");
            emitPush(FP, "backup parent fp.");
            backend.emitADDI(FP, SP, 2 * backend.getWordSize(), "Change FP to apply new frame.");

            // jal will set RA register properly.
            // FIXME: Testing fucking silly $print
            backend.emitADDI(SP, SP, 2 * backend.getWordSize(), "FUCKING SILLY BUILTIN CODE");
            backend.emitJAL(calledLabel, "Call it!");

            backend.emitADDI(SP, FP, -2 * backend.getWordSize(), "Revert all local variables on this dying frame.");
            emitPop(FP, "restore parent FP.");
            emitPop(RA, "restore parent return address.");
        }
        public void emitRet(String comment) {
            // All local variable should already be poped.
            // Return value should be in A0
            emitNoop(comment);
            backend.emitJR(RA, "Return to caller.");
        }
        public void emitPushIntVal(RiscVBackend.Register tmpReg, Integer val, String comment) {
            emitNoop(comment);
            backend.emitLI(tmpReg, val, "INT VAL");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 0, "OBJECT HEAD - DISPATCH TBL = NULL");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 4, "OBJECT HEAD - SIZE = 3+1");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 1, "OBJECT HEAD - TYPE = INT");
            emitPush(tmpReg, null);
            backend.emitMV(tmpReg, SP, "Return INT address.");
        }
        public void emitPushBoolVal(RiscVBackend.Register tmpReg, Boolean val, String comment) {
            emitNoop(comment);
            backend.emitLI(tmpReg, val ? 1 : 0, "BOOL VAL");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 0, "OBJECT HEAD - DISPATCH TBL = NULL");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 4, "OBJECT HEAD - SIZE = 3+1");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 2, "OBJECT HEAD - TYPE = BOOL");
            emitPush(tmpReg, null);
            backend.emitMV(tmpReg, SP, "Return BOOL address.");
        }
        public void emitPushStrVal(RiscVBackend.Register tmpReg, String val, String comment) {
            emitNoop(comment);
            byte [] bytes = val.getBytes(StandardCharsets.US_ASCII);
            int dataLen = bytes.length / backend.getWordSize() + 1; // Equals to ceil((len+1) DIVIDE WORD_SIZE)

            for(int cter = 0; cter < dataLen; ++cter) {
                int myRangeBegin = (dataLen-cter-1) * backend.getWordSize();
                int myRangeEnd = Math.min(myRangeBegin + backend.getWordSize(), bytes.length);
                int curr = 0;
                assert backend.getWordSize() == 4; // curr should be 4 byte = int
                String debug_pushed_str = "";
                for(int shift = 0; shift < myRangeEnd - myRangeBegin; ++shift) {
                    curr += bytes[myRangeBegin + shift] << (shift * 8);
                    debug_pushed_str += Integer.toString(bytes[myRangeBegin + shift]) + " ";
                }

                backend.emitLI(tmpReg, curr, "STR VAL - " + debug_pushed_str);
                emitPush(tmpReg, "STR VAL - PUSH A BLOCK");
            }

            backend.emitLI(tmpReg, bytes.length, "STR ATTR - __len__");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 0, "OBJECT HEAD - DISPATCH TBL = NULL");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 3 + dataLen + 1, "OBJECT HEAD - SIZE");
            emitPush(tmpReg, null);
            backend.emitLI(tmpReg, 3, "OBJECT HEAD - TYPE = STR");
            emitPush(tmpReg, null);
            backend.emitMV(tmpReg, SP, "Return STR OBJ address.");
        }

    }

    // Every function needs a register manager.
    // Every variable should borrow a register, and maybe it's ok to never return it.
    //     because our toy compiler will only be tested on naive program.
    //     manager won't swap register to stack smartly (with graph coloring technic),
    //     because it doesn't worth my effort(nobody pay for that).
    private class registerManager {
        private Map<RiscVBackend.Register, Boolean> usageMap; // true if the reg is in use.
        registerManager() {
            usageMap = new LinkedHashMap<>();
            usageMap.put(T0, false);
            usageMap.put(T1, false);
            usageMap.put(T2, false);
            usageMap.put(T3, false);
            usageMap.put(T4, false);
            usageMap.put(T5, false);
            usageMap.put(T6, false);
            usageMap.put(A1, false);
            usageMap.put(A2, false);
            usageMap.put(A3, false);
            usageMap.put(A4, false);
            usageMap.put(A5, false);
            usageMap.put(A6, false);
            usageMap.put(A7, false);
            usageMap.put(A0, false);
        }

        public RiscVBackend.Register borrowOne() {
            for(RiscVBackend.Register reg : usageMap.keySet()) {
                if(usageMap.put(reg, true) == false) {
                    return reg;
                }
            }
            throw new RuntimeException("Too many variable, not enough register.");
        }

        public void returnOne(RiscVBackend.Register reg) {
            boolean old = usageMap.put(reg, false);
            assert old == true;
        }
    }

    /** A code generator emitting instructions to BACKEND. */
    public CodeGenImpl(RiscVBackend backend) {
        super(backend);
        betterBackend = new BetterRsicVBackend(backend);
    }

    private BetterRsicVBackend betterBackend;

    /** Operation on None. */
    private final Label errorNone = new Label("error.None");
    /** Division by zero. */
    private final Label errorDiv = new Label("error.Div");
    /** Index out of bounds. */
    private final Label errorOob = new Label("error.OOB");

    /**
     * Emits the top level of the program.
     *
     * This method is invoked exactly once, and is surrounded
     * by some boilerplate code that: (1) initializes the heap
     * before the top-level begins and (2) exits after the top-level
     * ends.
     *
     * You only need to generate code for statements.
     *
     * @param statements top level statements
     */
    protected void emitTopLevel(List<Stmt> statements) {
        StmtAnalyzer stmtAnalyzer = new StmtAnalyzer(null);
        backend.emitADDI(SP, SP, -2 * backend.getWordSize(),
                         "Saved FP and saved RA (unused at top level).");
        backend.emitSW(ZERO, SP, 0, "Top saved FP is 0.");
        backend.emitSW(ZERO, SP, 4, "Top saved RA is 0.");
        backend.emitADDI(FP, SP, 2 * backend.getWordSize(),
                         "Set FP to previous SP.");

        for (Stmt stmt : statements) {
            stmt.dispatch(stmtAnalyzer);
        }
        backend.emitLI(A0, EXIT_ECALL, "Code for ecall: exit");
        backend.emitEcall(null);
    }

    /**
     * Emits the code for a function described by FUNCINFO.
     *
     * This method is invoked once per function and method definition.
     * At the code generation stage, nested functions are emitted as
     * separate functions of their own. So if function `bar` is nested within
     * function `foo`, you only emit `foo`'s code for `foo` and only emit
     * `bar`'s code for `bar`.
     */
    protected void emitUserDefinedFunction(FuncInfo funcInfo) {
        backend.emitGlobalLabel(funcInfo.getCodeLabel());
        StmtAnalyzer stmtAnalyzer = new StmtAnalyzer(funcInfo);

        for (Stmt stmt : funcInfo.getStatements()) {
            stmt.dispatch(stmtAnalyzer);
        }

        backend.emitMV(A0, ZERO, "Returning None implicitly");
        backend.emitLocalLabel(stmtAnalyzer.epilogue, "Epilogue");

        // FIXME: {... reset fp etc. ...}
        backend.emitJR(RA, "Return to caller");
    }

    /** An analyzer that encapsulates code generation for statments. */
    private class StmtAnalyzer extends AbstractNodeAnalyzer<RiscVBackend.Register> {
        /*
         * The symbol table has all the info you need to determine
         * what a given identifier 'x' in the current scope is. You can
         * use it as follows:
         *   SymbolInfo x = sym.get("x");
         *
         * A SymbolInfo can be one the following:
         * - ClassInfo: a descriptor for classes
         * - FuncInfo: a descriptor for functions/methods
         * - AttrInfo: a descriptor for attributes
         * - GlobalVarInfo: a descriptor for global variables
         * - StackVarInfo: a descriptor for variables allocated on the stack,
         *      such as locals and parameters
         *
         * Since the input program is assumed to be semantically
         * valid and well-typed at this stage, you can always assume that
         * the symbol table contains valid information. For example, in
         * an expression `foo()` you KNOW that sym.get("foo") will either be
         * a FuncInfo or ClassInfo, but not any of the other infos
         * and never null.
         *
         * The symbol table in funcInfo has already been populated in
         * the base class: CodeGenBase. You do not need to add anything to
         * the symbol table. Simply query it with an identifier name to
         * get a descriptor for a function, class, variable, etc.
         *
         * The symbol table also maps nonlocal and global vars, so you
         * only need to lookup one symbol table and it will fetch the
         * appropriate info for the var that is currently in scope.
         */

        /** Symbol table for my statements. */
        private SymbolTable<SymbolInfo> sym;

        /** Label of code that exits from procedure. */
        protected Label epilogue;

        /** The descriptor for the current function, or null at the top
         *  level. */
        private FuncInfo funcInfo;

        private registerManager regMgr = new registerManager();

        /** An analyzer for the function described by FUNCINFO0, which is null
         *  for the top level. */
        StmtAnalyzer(FuncInfo funcInfo0) {
            funcInfo = funcInfo0;
            if (funcInfo == null) {
                sym = globalSymbols;
            } else {
                sym = funcInfo.getSymbolTable();
            }
            epilogue = generateLocalLabel();
        }

        // FIXME: Example of statement.
        @Override
        public RiscVBackend.Register analyze(ReturnStmt node) {
            // FIXME: Here, we emit an instruction that does nothing. Clearly,
            // this is wrong, and you'll have to fix it.
            // This is here just to demonstrate how to emit a
            // RISC-V instruction.
            backend.emitMV(ZERO, ZERO, "No-op");
            return null;
        }

        // FIXME: More, of course.
        @Override
        public RiscVBackend.Register analyze(ExprStmt node) {
            node.expr.dispatch(this);
            return null;
        }

        @Override
        public RiscVBackend.Register analyze(CallExpr node) {
            SymbolInfo called = sym.get(node.function.name);
            if(called instanceof FuncInfo) {
                FuncInfo func = (FuncInfo) called;
                // TODO: Push the FUCKING arg0 (outter function frame ptr)
                List<RiscVBackend.Register> args_reg = new ArrayList<>();
                for(Expr expr : node.args) {
                    // TODO: sub-expr may use stack.
                    RiscVBackend.Register result = expr.dispatch(this);
                    if(result == null) {
                        throw new RuntimeException("NotImplemented: Expression " + expr.getClass().getName() + " returns null register.");
                    }
                    args_reg.add(result);
                }
                for(RiscVBackend.Register reg : args_reg) {
                    betterBackend.emitPush(reg, "Push function arguments.");
                    regMgr.returnOne(reg);
                }

                betterBackend.emitCall(func.getCodeLabel(), null);

                for(Object reg : args_reg)
                    betterBackend.emitPop(null, "Pop function arguments.");

                return A0; // Function return value is always in A0.
            }
            else if(called instanceof ClassInfo) {
                // TODO: Fuck the fucking framework.
            }
            return null;
        }

        @Override
        public RiscVBackend.Register analyze(BinaryExpr node) {
            RiscVBackend.Register leftRes = node.left.dispatch(this);
            RiscVBackend.Register rightRes = node.right.dispatch(this);
            if(leftRes == null || rightRes == null)
                throw new RuntimeException("NOT IMPL");
            switch(node.operator) {
                case "+":
                    if(node.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        RiscVBackend.Register savedLeftAddr = regMgr.borrowOne();
                        backend.emitMV(savedLeftAddr, leftRes, "Backup reg leftRes");
                        backend.emitLW(leftRes, leftRes, 3 * backend.getWordSize(), "Operator+ Fetch left int result");
                        backend.emitLW(rightRes, rightRes, 3 * backend.getWordSize(), "Operator+ Fetch right int result");
                        backend.emitADD(leftRes, leftRes, rightRes, "Operator+::INT");
                        regMgr.returnOne(rightRes);
                        backend.emitSW(leftRes, savedLeftAddr, 3 * backend.getWordSize(), "Operator+::INT save final res.");
                        regMgr.returnOne(leftRes);
                        return savedLeftAddr;
                    }
                    if(node.left.getInferredType().isListType()) {
                        // TODO: Merge two list.
                        throw new RuntimeException("NOTIMPLEMENTED: ListMerge");
                    }
                    if(node.left.getInferredType().equals(SymbolType.STR_TYPE)) {
                        // TODO: Merge two str.
                        throw new RuntimeException("NOTIMPLEMENTED: StrMerge");
                    }
                case "-":
                    if(node.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        RiscVBackend.Register savedLeftAddr = regMgr.borrowOne();
                        backend.emitMV(savedLeftAddr, leftRes, "Backup reg leftRes");
                        backend.emitLW(leftRes, leftRes, 3 * backend.getWordSize(), "Operator- Fetch left int result");
                        backend.emitLW(rightRes, rightRes, 3 * backend.getWordSize(), "Operator- Fetch right int result");
                        backend.emitSUB(leftRes, leftRes, rightRes, "Operator-::INT");
                        regMgr.returnOne(rightRes);
                        backend.emitSW(leftRes, savedLeftAddr, 3 * backend.getWordSize(), "Operator-::INT save final res.");
                        regMgr.returnOne(leftRes);
                        return savedLeftAddr;
                    }
                case "*":
                    if(node.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        RiscVBackend.Register savedLeftAddr = regMgr.borrowOne();
                        backend.emitMV(savedLeftAddr, leftRes, "Backup reg leftRes");
                        backend.emitLW(leftRes, leftRes, 3 * backend.getWordSize(), "Operator* Fetch left int result");
                        backend.emitLW(rightRes, rightRes, 3 * backend.getWordSize(), "Operator* Fetch right int result");
                        backend.emitMUL(leftRes, leftRes, rightRes, "Operator*::INT");
                        regMgr.returnOne(rightRes);
                        backend.emitSW(leftRes, savedLeftAddr, 3 * backend.getWordSize(), "Operator*::INT save final res.");
                        regMgr.returnOne(leftRes);
                        return savedLeftAddr;
                    }
            }
            return null;
        }

        @Override
        public RiscVBackend.Register analyze(IntegerLiteral node) {
            // emitConstant(node, ValueType.INT_TYPE, "Set constant int literal.");
            RiscVBackend.Register tmpReg = regMgr.borrowOne();
            betterBackend.emitPushIntVal(tmpReg, node.value, "Push int literal");
            return tmpReg;
        }

        @Override
        public RiscVBackend.Register analyze(BooleanLiteral node) {
            RiscVBackend.Register tmpReg = regMgr.borrowOne();
            betterBackend.emitPushBoolVal(tmpReg, node.value, "Push bool literal");
            return tmpReg;
        }

        @Override
        public RiscVBackend.Register analyze(StringLiteral node) {
            RiscVBackend.Register tmpReg = regMgr.borrowOne();
            betterBackend.emitPushStrVal(tmpReg, node.value, "push string literal.");
            return tmpReg;
        }
    }

    /**
     * Emits custom code in the CODE segment.
     *
     * This method is called after emitting the top level and the
     * function bodies for each function.
     *
     * You can use this method to emit anything you want outside of the
     * top level or functions, e.g. custom routines that you may want to
     * call from within your code to do common tasks. This is not strictly
     * needed. You might not modify this at all and still complete
     * the assignment.
     *
     * To start you off, here is an implementation of three routines that
     * will be commonly needed from within the code you will generate
     * for statements.
     *
     * The routines are error handlers for operations on None, index out
     * of bounds, and division by zero. They never return to their caller.
     * Just jump to one of these routines to throw an error and
     * exit the program. For example, to throw an OOB error:
     *   backend.emitJ(errorOob, "Go to out-of-bounds error and abort");
     *
     */
    protected void emitCustomCode() {
        emitErrorFunc(errorNone, "Operation on None");
        emitErrorFunc(errorDiv, "Divison by zero");
        emitErrorFunc(errorOob, "Index out of bounds");
    }

    /** Emit an error routine labeled ERRLABEL that aborts with message MSG. */
    private void emitErrorFunc(Label errLabel, String msg) {
        backend.emitGlobalLabel(errLabel);
        backend.emitLI(A0, ERROR_NONE, "Exit code for: " + msg);
        backend.emitLA(A1, constants.getStrConstant(msg),
                       "Load error message as str");
        backend.emitADDI(A1, A1, getAttrOffset(strClass, "__str__"),
                         "Load address of attribute __str__");
        backend.emitJ(abortLabel, "Abort");
    }
}
